#include <QApplication>
#include <QWidget>
#include <QCheckBox>
#include <QPushButton>
#include <QDebug>
#include <QVBoxLayout>
#include <QMainWindow>
#include <QAction>
#include <QMenu>
#include <QMenuBar>
#include <QToolBar>
#include <QLabel>
#include <QDesktopWidget>
#include <QMoveEvent>
#if QT_VERSION >= 0x050000
#  include <QScreen>
#  include <QWindow>
#endif

bool optNativeChild = false;
bool optReportMove = false;
bool optVerboseDumpNonNativeWidgets = false;

static bool hasNativeChild(const QObjectList &ol)
{
    foreach (const QObject *co, ol)
        if (co->isWidgetType())
            if (static_cast<const QWidget *>(co)->internalWinId())
                return true;
    return false;
}

static void dumpWidgetRecursion(QTextStream &str, const QWidget *w, int depth = 0)
{
    const QRect geom = w->geometry();
    const bool native = w->internalWinId();
    QObjectList children = w->children();
    if (optVerboseDumpNonNativeWidgets || native || hasNativeChild(children)) {
        const QPoint globalPos = w->isWindow() ? geom.topLeft() : w->mapToGlobal(geom.topLeft());
        if (depth)
            str << QString(depth * 2, QLatin1Char(' '));
        str << '"' << w->metaObject()->className() << "\"/\"" << w->objectName() << "\" "
            << (w->isVisible() ? "[visible] " : "[hidden] ");
        if (native)
            str << "[native] ";
        if (native || optVerboseDumpNonNativeWidgets) {
            if (w->isWindow())
                str << "[top] ";
            str << "f=0x" << hex << unsigned(w->windowFlags()) << dec
                << "\" " << geom.width() << 'x' << geom.height()
                << forcesign << geom.x() << geom.y()
                << " Global: " << globalPos.x() << globalPos.y() << noforcesign
                << " screen #" << QApplication::desktop()->screenNumber(w);
#if QT_VERSION >= 0x050000
            if (const QWindow *qw = w->windowHandle()) {
                QScreen *scr = qw->screen();
                str << " '" << scr->name() << "' #" << QGuiApplication::screens().indexOf(scr);
            } else {
                str << " no QWindow";
            }
        }
#endif
        str << '\n';
    }
    foreach (const QObject *co, children) {
        if (co->isWidgetType()) {
            const QWidget *cw = static_cast<const QWidget *>(co);
            dumpWidgetRecursion(str, cw, depth + 1);
        }
    }
}

static void dumpWidgets()
{
    QString d;
    QTextStream str(&d);
    foreach (QWidget *tw, QApplication::topLevelWidgets())
        dumpWidgetRecursion(str, tw);
    qDebug().nospace() << d;
}

#if QT_VERSION >= 0x050000
QDebug operator<<(QDebug d, const QScreen &scr)
{
    d << scr.name() << scr.geometry() << "Phys:" << scr.physicalSize()
      <<  "Logical DPI:" << scr.logicalDotsPerInchX() << 'x' << scr.logicalDotsPerInchY()
      << "Physical DPI:" << scr.physicalDotsPerInchX() << 'x' << scr.physicalDotsPerInchY();
    return d;
}
#endif

class Widget : public QMainWindow
{
    Q_OBJECT
public:
    explicit Widget(int n, QWidget *parent = 0);

public slots:
    void tileOthers();
    void dump() { dumpWidgets(); }
    void dumpScreens();
#if QT_VERSION >= 0x050000
    void screenAdded(QScreen *);
    void qWindowScreenChanged(QScreen *);
#endif
    void screenDestroyed();

protected:
    void moveEvent(QMoveEvent *);
};

Widget::Widget(int nn, QWidget *parent) : QMainWindow(parent)
{
    const QString n = QString::number(nn + 1);
    setWindowTitle(QT_VERSION_STR + QString(" #") + n);
    setObjectName("Widget" + n);

    QMenuBar *mb =menuBar();
    mb->setObjectName("MenuBar" + n);
    QMenu *fm = mb->addMenu("File");

    fm->setObjectName("Menu" + n);
    QToolBar *t = addToolBar("File");
    t->setObjectName("ToolBar" + n);

    QAction *a = new QAction("Dump Widgets", this);
    a->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_D));
    connect(a, SIGNAL(triggered()), this, SLOT(dump()));
    fm->addAction(a);
    t->addAction(a);

    a = new QAction("Dump screens", this);
    a->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_S));
    connect(a, SIGNAL(triggered()), this, SLOT(dumpScreens()));
    fm->addAction(a);

    a = new QAction("Tile Others", this);
    a->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_I));
    connect(a, SIGNAL(triggered()), this, SLOT(tileOthers()));
    fm->addAction(a);
    fm->addSeparator();
    a = new QAction("Show full screen", this);
    a->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_F));
    connect(a, SIGNAL(triggered()), this, SLOT(showFullScreen()));
    fm->addAction(a);
    a = new QAction("Show minimized", this);
    a->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_M));
    connect(a, SIGNAL(triggered()), this, SLOT(showMinimized()));
    fm->addAction(a);
    a = new QAction("Show normal", this);
    a->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_N));
    connect(a, SIGNAL(triggered()), this, SLOT(showNormal()));
    fm->addAction(a);
    fm->addSeparator();

    a = new QAction("Quit", this);
    a->setShortcut(QKeySequence(Qt::CTRL + Qt::Key_Q));
    connect(a, SIGNAL(triggered()), qApp, SLOT(quit()));
    fm->addAction(a);
    t->addAction(a);

#if QT_VERSION >= 0x050000
    connect(qApp, SIGNAL(screenAdded(QScreen*)), this, SLOT(screenAdded(QScreen*)));
    foreach (QScreen *s, QGuiApplication::screens())
        connect(s, SIGNAL(destroyed()), this, SLOT(screenDestroyed()));
#endif

    QLabel *label = new QLabel(optNativeChild ? "Native Central" : "Central");
    label->setObjectName("Label" + n);
    setCentralWidget(label);
    if (optNativeChild)
        label->winId();
}

void Widget::dumpScreens()
{
#if QT_VERSION >= 0x050000
    static QList<QScreen *> screens = QGuiApplication::screens();
    for (int s = 0; s < screens.size(); ++s)
        qDebug() << '#' << s << screens.at(s) << *screens.at(s);
#endif
}

void Widget::screenDestroyed()
{
    QString name;
    if (QScreen *s = qobject_cast<QScreen *>(sender()))
        name = s->name();

    qDebug() << __FUNCTION__ << this << sender() << name;
}

void Widget::moveEvent(QMoveEvent *me)
{
    if (optReportMove)
        qDebug() << __FUNCTION__ << this << me->oldPos() << me->pos() << me->spontaneous();
}

#if QT_VERSION >= 0x050000
void Widget::screenAdded(QScreen *s)
{
    qDebug() << __FUNCTION__ << this << s << *s;
    connect(s, SIGNAL(destroyed()), this, SLOT(screenDestroyed()));
}

void Widget::qWindowScreenChanged(QScreen *s)
{
    static int n = 0;
    qDebug() << __FUNCTION__ << '#' << n++ << sender() << s->name();
}
#endif

void Widget::tileOthers()
{
    QWidget *last = this;
    foreach (QWidget *w, QApplication::topLevelWidgets()) {
        if (w != this && w->windowType() == Qt::Window ) {
            w->move(last->frameGeometry().topRight() + QPoint(50, 0));
            last = w;
        }
    }
}

int main(int argc, char *argv[])
{
    typedef QSharedPointer<Widget> WidgetPtr;
    QApplication a(argc, argv);
    qDebug() << QT_VERSION_STR;
    QStringList args = QCoreApplication::arguments();
    optNativeChild = args.contains("-w");
    optReportMove = args.contains("-m");
    optVerboseDumpNonNativeWidgets = args.contains("-v");
    int n = 2;
    int index = args.indexOf("-n");
    if (index > 0)
        n = args.at(index + 1).toInt();

    int y = ((QT_VERSION >> 16) - 4) * 200 + 200;
    QList<WidgetPtr> widgets;
    for (int i = 0; i < n; ++i) {
        WidgetPtr w(new Widget(i));
        w->move(300 + 300 * i, y);
        w->show();
        widgets.append(w);
    }
#if QT_VERSION >= 0x050000
    QWidget *reportWidget = widgets.front().data();
    foreach (QWindow *w, QGuiApplication::allWindows())
        QObject::connect(w, SIGNAL(screenChanged(QScreen*)),
                         reportWidget, SLOT(qWindowScreenChanged(QScreen*)));
#endif
    return a.exec();
}

#include "main.moc"
